/*****************************************************/
/*                       sga.c                       */
/*　SGA 설명용 프로그램  　    　                    */
/* 사용방법　                                        */
/*       $./sga                                      */
/*****************************************************/

/* 헤더 파일 포함 */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

/* 기호 상수의 정의 */
#define DATANO 20   /* 데이터 수 */
#define POOLSIZE 30 /* 풀 사이즈 */
#define LASTG 200    /* 중단 세대*/
#define MRATE 0.01  /* 돌연변이 확률 */
#define SEED 65535    /* 난수 시드 */
#define YES 1   /* yes에 대응하는 정수 */
#define NO 0    /* no에 대응하는 정수 */
#define VALUE 499 /* 목표치 */
#define MAXFIT 1000    /* 적응도 계산에 이용할 수치 */

/* 함수의 프로토타입 선언 */
int evalfit(int gene[]) ; /*  적응도 계산 */
void mating(int pool[POOLSIZE][DATANO]) ; /* 교차 */
void mutation(int pool[POOLSIZE][DATANO]) ; /*  돌연변이 */
void printp(int pool[POOLSIZE][DATANO]) ; /* 결과출력 */
void initpool(int pool[POOLSIZE][DATANO]) ;
                         /* 초기 집단 생성 */
int rndn() ; /* n미만의 난수 생성 */
int notval(int v) ;/* 진리값 반전 */
int select(int roulette[POOLSIZE],int totalfitness) ;
                    /* 선택 */
void copypool(int pool[POOLSIZE][DATANO]
          ,int nextpool[POOLSIZE][DATANO]) ;
                 /* 다음 세대의 풀을 복사 */
void crossing(int m[],int p[],int c1[],int c2[]) ;
                /* 특정한 두 유전자의 교차 */

/* 문제의 수치를 설정 */
int q[]={31,41,59,26,53,58,97,93,23,84,
         -62,-64,-33,-83,-27,-95,-2,-88,-41,-97} ;

/*******************/
/*    main()함수   */
/*******************/
int main(int argc,char *argv[])
{
 int pool[POOLSIZE][DATANO] ; /* 유전자 풀 */
 int generation;/* 현재 세대수 */

 /* 난수 초기화 */
 srand(SEED) ;

 /* 초기집단 생성*/
 initpool(pool) ;

 /* 중단 세대까지 반복 */
 for(generation=0;generation<LASTG;++generation){
  printf("%d세대\n",generation) ;
  mating(pool) ;/* 교차 */
  mutation(pool) ;/* 돌연변이 */
  printp(pool) ;/* 결과출력 */
 }
 
 return 0 ;
}


/************************/
/*   select()함수       */
/*       선택           */
/************************/
int select(int roulette[POOLSIZE],int totalfitness)
{
 int i ;/* 반복 제어 문자 */
 int ball ;/* 공（선택 위치의 수치）*/
 int acc=0 ;/* 적응도 적산값 */

 ball=rndn(totalfitness) ;
 for(i=0;i<POOLSIZE;++i){
  acc+=roulette[i] ;/* 적응도 적선 */
  if(acc>ball) break ;/* 대응하는 유전자 */
 }
   
 return i ;

}


/************************/
/*   mating()함수       */
/*        교차          */
/************************/
void mating(int pool[POOLSIZE][DATANO])
{
 int i ;/* 반복 제어 변수 */
 int totalfitness=0 ;/* 적응도의 합계 */
 int nextpool[POOLSIZE][DATANO] ;/* 자식 세대 */
 int roulette[POOLSIZE] ;/* 적응도 저장 */
 int mama,papa ;/* 부모 유전자 번호 */

 /* 룰렛 작성 */
 for(i=0;i<POOLSIZE;++i){
  roulette[i]=evalfit(pool[i]) ;
  /* 적응도의 합계를 계산 */
  totalfitness+=roulette[i] ;
 }

 /* 선택과 교차를 반복한다 */
 for(i=0;i<POOLSIZE/2;++i){
  do{/* 부모 선택 */
   mama=select(roulette,totalfitness) ;
   papa=select(roulette,totalfitness) ;
  }while(mama==papa) ;/*  중복제거 */
  /* 특정한 두 유전자의 교차 */
  crossing(pool[mama],pool[papa]
            ,nextpool[i*2],nextpool[i*2+1])  ;
 }

 /* 다음 세대 풀을 복사 */
 copypool(pool,nextpool) ;

}
/************************/
/*  crossing() 함수     */
/* 특정한 두 유전의 교차*/
/************************/
void crossing(int m[],int p[],int c1[],int c2[])
{
 int j ;/* 반복 제어 변수 */
 int cp ;/*교차할 점 */

 /* 교차점 결정 */
 cp =rndn(DATANO) ;

 /* 유전자의 전반 부분 복사 */
 for(j=0;j<cp;++j){
  c1[j]=m[j] ;
  c2[j]=p[j] ;
 }
 
 /*유전자의 후반 부분 복사*/
 for(;j<DATANO;++j){
  c2[j]=m[j] ;
  c1[j]=p[j] ;
 }
}

/************************/
/*  copypool() 함수     */
/* 다음 세대의 풀을 복사*/
/************************/
void copypool(int pool[POOLSIZE][DATANO]
          ,int nextpool[POOLSIZE][DATANO])
{
 int i,j ;/* 반복 제어 변수 */

 for(i=0;i<POOLSIZE;++i)
  for(j=0;j<DATANO;++j)
   pool[i][j]=nextpool[i][j] ;
}

/************************/
/*   evalfit()함수      */
/*   적응도 계산        */
/************************/
int evalfit(int g[])
{
 int i ;/* 반복 제어 변수 */
 int fitness=0 ;/* 적응도 도중값 */

 for(i=0;i<DATANO;++i)
  fitness+=g[i]*q[i] ;

 return MAXFIT-abs(VALUE-fitness) ;
}

/***********************/
/*   printp() 함수     */
/*   출력결과          */
/***********************/
void printp(int pool[POOLSIZE][DATANO])
{
 int i,j ;/* 반복 제어 변수 */
 int fitness ;/* 적응도 */
 double totalfitness=0 ;/* 적응도의 합계 */
 int elite,bestfit=0 ;/*엘리트 유전자 처리용 변수 */ 

 for(i=0;i<POOLSIZE;++i){
  for(j=0;j<DATANO;++j)
   printf("%1d",pool[i][j]) ;
  fitness=evalfit(pool[i]) ;
  printf("\t%d\n",fitness) ;
  if(fitness>bestfit){/* 엘리트 해 */
   bestfit=fitness ;
   elite=i ;
  }
  totalfitness+=fitness ;
 }
 /* 엘리트 해의 적응도 출력 */
 printf("%d\t%d   ",elite,bestfit) ;
 /* 평균 적응도 출력 */
 printf("%lf\n",totalfitness/POOLSIZE) ;
}


/***********************/
/*   initpool()함수    */
/*   초기집단 생성     */
/***********************/
void initpool(int pool[POOLSIZE][DATANO])
{
 int i,j ;/* 반복 제어 변수 */

 for(i=0;i<POOLSIZE;++i)
  for(j=0;j<DATANO;++j)
   pool[i][j]=rndn(2) ;

}


/************************/
/*     rndn()함수       */
/*    n미만의 난수 생성 */
/************************/
int rndn(int l)
{
 int rndno ;/* 생성한 난수 */

 while((rndno=((double)rand()/RAND_MAX)*l)==l) ;

 return rndno;
}

/***********************/
/*   mutation()함수    */
/*   돌연변이          */
/***********************/
void mutation(int pool[POOLSIZE][DATANO])
{
 int i,j ;/* 반복 제어 변수 */

 for(i=0;i<POOLSIZE;++i)
  for(j=0;j<DATANO;++j)
   if((double)rndn(100)/100.0<=MRATE)
    /* 반전의 돌연변이 */
    pool[i][j]=notval(pool[i][j]) ;

}


/************************/
/*   notval()함수       */
/*   진리값 반전        */
/************************/
int notval(int v)
{
 if(v==YES) return NO ;
 else return YES ;
}

